/******************************************
 * [d7DevServer]
 * Copyright (c) 2024 Kougen.Sai
 * Released under the MIT license
 * https://opensource.org/licenses/mit-license.php 
 ******************************************/
package com.cycau.d7server.config;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Path;
import java.nio.file.Paths;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.Resource;
import org.springframework.web.servlet.config.annotation.ResourceHandlerRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer;
import org.springframework.web.servlet.resource.PathResourceResolver;

@Configuration
public class MvcConfig implements WebMvcConfigurer {
	@Value("${d7server.resource.external}")
	private String externalPath;
	@Value("${d7server.resource.index}")
	private String resourceIndex;
	@Value("${d7server.resource.mockfolder}")
	private String mockfolder;

	private static final Logger log = LoggerFactory.getLogger("DDDDDDD");

	@Override
	public void addResourceHandlers(ResourceHandlerRegistry registry) {
		if (!externalPath.isEmpty()) {
			if (externalPath.endsWith("/")) {
				externalPath.substring(0, externalPath.length()-1);
			}
			if (externalPath.startsWith("./")) {
				Path path = Paths.get("");
				externalPath = path.toAbsolutePath() + externalPath.substring(1);
			}
		}

		if (resourceIndex.isEmpty()) {
			resourceIndex = "index.html";
		} else if (resourceIndex.startsWith("/")){
			resourceIndex = resourceIndex.substring(1);
		}

		if (mockfolder.startsWith("/")){
			mockfolder = mockfolder.substring(1);
		}

		registry
		.addResourceHandler("/**")
		.addResourceLocations("classpath:/public/")
		.setCachePeriod(60*60*24)
		.resourceChain(true)
		.addResolver(externalPath.isEmpty() ? new SysResourceResolver() : new ExternalResourceResolver(externalPath));
		
		if (externalPath.isEmpty()) {
			log.info("resource folder: JAR/public");
			log.info("index file     : JAR/public/{}", resourceIndex);
			log.info("mock folder    : JAR/public/{}", mockfolder);
		} else {
			log.info("resource folder: {}", externalPath);
			log.info("index file     : {}/{}", externalPath ,resourceIndex);
			log.info("mock folder    : {}/{}", externalPath ,mockfolder);
		}
	}

	public class SysResourceResolver extends PathResourceResolver {
		@Override
		protected Resource getResource(String resourcePath, Resource location) throws IOException {
			if (resourcePath.indexOf(".") > 0) {
				Resource resource = super.getResource(resourcePath, location);;
				if (resource == null) {
					log.error("Not found file. JAR/public/{}", resourcePath);
				}
				return resource;
			}

			if (!resourceIndex.isEmpty()) {
				Resource resource = super.getResource(resourceIndex, location);;
				if (resource == null) {
					log.error("Not found App index file. JAR/public/{}", resourceIndex);
				}
				return resource;
			}

			// Dead code. find /resourcePath/index.xxx
			InputStream is = Thread.currentThread().getContextClassLoader().getResourceAsStream("public/" + resourcePath);
			if (is == null) is = ClassLoader.getSystemClassLoader().getResourceAsStream(resourcePath);
			if (is == null) {
				log.error("Not found folder. JAR/public/{}", resourcePath);
				return null;
			}

			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			String fileName = null;
			while ((fileName = br.readLine()) != null) {
				if (fileName.startsWith("index.")) break;
			}
			br.close();
			is.close();

			if (fileName == null) return null;
			return super.getResource(resourcePath + fileName, location);
		}
	}

	public class ExternalResourceResolver extends PathResourceResolver {
		private URLClassLoader exResourceLoader;

		public ExternalResourceResolver(String resourceRoot) {
			try {
				exResourceLoader = new URLClassLoader(new URL[]{new File(resourceRoot).toURI().toURL()});
			} catch (Exception e) {
				e.printStackTrace();
				throw new RuntimeException("[DevServer] when loading external resource.");
			}
		}

		@Override
		protected Resource getResource(String resourcePath, Resource location) throws IOException {
			if (resourcePath.indexOf(".") > 0) {
				Resource resource = new ClassPathResource(resourcePath, this.exResourceLoader);
				if (resource.isReadable()) return resource;

				log.error("Not found file. {}/{}", externalPath , resourcePath);
				return null;
			}

			if (!resourceIndex.isEmpty()) {
				Resource resource = new ClassPathResource(resourceIndex, this.exResourceLoader);
				if (resource.isReadable()) return resource;

				log.error("Not found App index file. {}/{}", externalPath, resourceIndex);
				return null;
			}

			// Dead code. find /resourcePath/index.xxx
			InputStream is = this.exResourceLoader.getResourceAsStream(resourcePath);
			if (is == null) {
				log.error("Not found folder. {}/{}", externalPath, resourcePath);
				return null;
			}

			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			String fileName = null;
			while ((fileName = br.readLine()) != null) {
				if (fileName.startsWith("index.")) break;
			}
			br.close();
			is.close();

			if (fileName == null) return null;
			return new ClassPathResource(externalPath + "/" + resourcePath + fileName, this.exResourceLoader);
		}
	}
}