/******************************************
 * [d7DevServer]
 * Copyright (c) 2024 Kougen.Sai
 * Released under the MIT license
 * https://opensource.org/licenses/mit-license.php 
 ******************************************/
package com.cycau.d7server.controller;

import java.util.Map;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.ArrayList;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Component;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.yaml.snakeyaml.Yaml;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;

import javax.annotation.PostConstruct;

@Component
public class MockData {
    @Value("${d7server.resource.external}")
    private String externalPath;
    @Value("${d7server.resource.mockfolder}")
    private String mockfolder;
    private ResourcePatternResolver resourcePatResolver;
    private static final Logger log = LoggerFactory.getLogger("DDDDDDD");

    @PostConstruct
    private void init() throws MalformedURLException {
    	if (!externalPath.isEmpty()) {
    		if (externalPath.endsWith("/")) {
    			externalPath.substring(0, externalPath.length()-1);
    		}
			if (externalPath.startsWith("./")) {
				Path path = Paths.get("");
				externalPath = path.toAbsolutePath() + externalPath.substring(1);
			}
    	}

    	if (mockfolder.startsWith("/")){
			mockfolder = mockfolder.substring(1);
		}

    	resourcePatResolver = new PathMatchingResourcePatternResolver();
    }

    @SuppressWarnings("unchecked")
    public ResponseEntity<Object> getResponse(String method, String uri, Map<String, Object> requestData) throws IOException {
    	method = method.toUpperCase();
        uri = mockfolder + uri;
    	
    	List<Map<String, Object>> mockData;
    	if (externalPath.isEmpty()) {
    		mockData = collectMockDataJAR(method, uri);
    	} else {
    		mockData = collectMockDataEXT(method, uri);
    	}

        for (Map<String, Object> mock : mockData) {
            Map<String, Object> mockRequest = (Map<String, Object>)mock.get("request");
            if (!equals(requestData.get("query"), mockRequest.get("query"), true)) {
            	log.info("Not matched with Query.  {}", mock.get("mock_name"));
            	continue;
            }
            if (!equals(requestData.get("header"), mockRequest.get("header"), false)) {
            	log.info("Not matched with Header. {}", mock.get("mock_name"));
            	continue;
            }
            if (!equals(requestData.get("data"), mockRequest.get("data"), true)) {
            	log.info("Not matched with Data.   {}", mock.get("mock_name"));
            	continue;
            }

            	log.info("Matched! with {}", mock.get("mock_name"));

            HttpStatus status = HttpStatus.OK;
            Map<String, Object> mockResponse = (Map<String, Object>)mock.get("response");
            if (mockResponse.containsKey("status")) status = HttpStatus.valueOf((int)mockResponse.get("status"));

            if (mockResponse.containsKey("header")) {
                MultiValueMap<String, String> headers = new LinkedMultiValueMap<>();
                Map<String, Object> mockHeader = (Map<String, Object>)mockResponse.get("header");
                if (mockHeader != null) {
                    for(Map.Entry<String, Object> keyVal : mockHeader.entrySet()) {
                        headers.add(keyVal.getKey(), keyVal.getValue() == null ? null : keyVal.getValue().toString());
                    }
                }

                return new ResponseEntity<>(mockResponse.get("body"), headers, status);
            }
            return new ResponseEntity<>(mockResponse.get("body"), status);
        }

        return new ResponseEntity<>("Not found mock data.", HttpStatus.NOT_FOUND);
    }

    private List<Map<String, Object>> collectMockDataJAR(String method, String uri) throws IOException {
    	uri = "public/" + uri;
    	List<Map<String, Object>> mocks = new ArrayList<>();
        
        Resource[] resources = null;
        try {
        	resources = resourcePatResolver.getResources("classpath:" + uri + "/*.*");
        } catch (Exception ex) {
        	log.error("Not found mock path. JAR/{}", uri);
        	return mocks;
        }

        Yaml yaml = new Yaml();
        for(Resource resource: resources) {
        	String tmp = resource.getFilename().toUpperCase();
            if (!tmp.startsWith(method)) continue;
            if (!tmp.endsWith(".YML")) continue;
			InputStream fileStream = resource.getInputStream();
            InputStreamReader fileReader = new java.io.InputStreamReader(fileStream, "UTF-8");
            try {
            	Map<String, Object> mockData = yaml.load(fileReader);
            	mockData.put("mock_name", "JAR/" + uri + "/" + resource.getFilename());
                mocks.add(mockData);
            } catch (Exception ex) {
            	log.error("Wrong yaml format. JAR/{}/{}", uri, resource.getFilename());
            }
            fileReader.close();
            fileStream.close();
		}

        if (mocks.size() < 1)
        	log.error("Not found mock file. JAR/{}/{}...yml", uri, method);
    
        return mocks;
    }
    private List<Map<String, Object>> collectMockDataEXT(String method, String uri) throws IOException {
    	List<Map<String, Object>> mocks = new ArrayList<>();
        
    	String resourcePath = uri;
    	URLClassLoader classLoader = new URLClassLoader(new URL[]{new File(externalPath).toURI().toURL()});
    	InputStream is = classLoader.getResourceAsStream(resourcePath);
        if (is == null) {
        	log.error("Not found mock path. {}/{}", externalPath, resourcePath);
        	classLoader.close();
        	return mocks;
        }

        BufferedReader br = new BufferedReader(new InputStreamReader(is, "UTF-8"));

        String fileName;
        Yaml yaml = new Yaml();
        while ((fileName = br.readLine()) != null) {
        	String tmp = fileName.toUpperCase();
            if (!tmp.startsWith(method)) continue;
            if (!tmp.endsWith(".YML")) continue;

            InputStream fileStream = classLoader.getResourceAsStream(resourcePath + "/" + fileName);
            InputStreamReader fileReader = new java.io.InputStreamReader(fileStream, "UTF-8");
            try {
            	Map<String, Object> mockData = yaml.load(fileReader);
            	mockData.put("mock_name", externalPath + "/" + resourcePath + "/" + fileName);
                mocks.add(mockData);
            } catch (Exception ex) {
            	log.error("Wrong yaml format. {}/{}/{}", externalPath, resourcePath, fileName);
            }
            fileReader.close();
            fileStream.close();
        }
        br.close();
        is.close();
    	classLoader.close();

        if (mocks.size() < 1)
        	log.error("Not found mock file. {}/{}/{}...yml", externalPath, resourcePath, method);
    
        return mocks;
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
	private static boolean equals(Object reqData, Object mock, boolean exactMatch) {
    	if ("*".equals(mock)) return true;
    	if ("**".equals(mock)) return true;

    	if (exactMatch) {
            if (mock == null || reqData == null) {
            	if (mock == null && reqData == null) return true;
            	return false;
            }
    	} else {
    		if (mock == null || reqData == null) {
        		if (mock == null) return true;
        		return false;
    		}
    	}

        if (mock instanceof String) return mock.equals(reqData);

        if (mock instanceof Map) {
            if (!(reqData instanceof Map)) return false;
            return compareMap((Map)reqData, (Map)mock, exactMatch);
        }
        if (mock instanceof List) {
            if (!(reqData instanceof List)) return false;
            return compareList((List)reqData, (List)mock, exactMatch);
        }

        return mock.toString().equals(reqData.toString());
    }
    private static boolean compareMap(Map<String, Object> reqData, Map<String, Object> mock, boolean exactMatch) {
    	if (exactMatch) {
            if (mock.size() != reqData.size()) return false;
    	} else {
            if (mock.size() > reqData.size()) return false;
    	}

        for(String key : mock.keySet()) {
            if (!reqData.containsKey(key)) return false;
            if (!equals(mock.get(key), reqData.get(key), true)) return false;
        }
        return true;
    }
    private static boolean compareList(List<Object> reqData, List<Object> mock, boolean exactMatch) {
    	if (exactMatch) {
            if (mock.size() != reqData.size()) return false;
    	} else {
            if (mock.size() > reqData.size()) return false;
    	}

        for(int idx=0; idx<mock.size(); idx++) {
            if (!equals(mock.get(idx), reqData.get(idx), true)) return false;
        }
        return true;
    }
}